/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: parser.cc                                    */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "parser.h"

// ******************************************
// ************** File parser ***************
// ******************************************

// Default constructor (common to all parsers)
Parser::Parser() {
   parseFile = NULL;
   parsedlines = readlines = parseFilePos = parseFileSize = 0UL;
   for (int i = 0; i<=6; i++) {
      fieldPos[i] = 0;
   }
   nowParsing = false;
   strcpy(curLine, "");
   strcpy(error, "");
}

// Destructor  (common to all parsers)
Parser::~Parser() {
   if (parseFile != NULL) {
      fclose(parseFile);
   }
}


/* Start parsing a given file; returns true if file access was successful;
   false otherwise (this also happens when trying to open a file while there
   is an already open one); this method is common to all parsers */
bool Parser::openFile(char *fileName) {
   struct stat fileStat;
   int parseFileDesc;
   if (parseFile == NULL) {
      parsedlines = 0;
      readlines = 0;
      strcpy(curLine, "");
      strcpy(error, "");
      nowParsing = false;
      parseFilePos = 0;
      for (int i = 0; i<=6; i++) {
         fieldPos[i] = 0;
      }
      parseFile = fopen(fileName, "r");
      if (parseFile != NULL) {
         parseFileDesc = fileno(parseFile);        /* Map file stream to
                                                      file descriptor (needed
                                                      to get file size) */
         fstat(parseFileDesc, &fileStat);          // Get file status (including size)
         parseFileSize = (unsigned long int) fileStat.st_size;
      }
      return (parseFile != NULL);
   }
   else {
      // There is a file that is already open
      return false;
   }
}

/* Assume input to be parsed comes from stdin; this method is common
   to all parsers */
void Parser::openStdin() {
   parsedlines = 0;
   readlines = 0;
   nowParsing = false;
   strcpy(curLine, "");
   strcpy(error, "");
   parseFilePos = 0;
   parseFileSize = 0;      // This will remain 0 until a new file is opened
   for (int i = 0; i<=6; i++) {
      fieldPos[i] = 0;
   }
   parseFile = stdin;
}

/* Stop parsing and close file (do not use it if working on stdin);
   this also resets counters (parsed lines, etc.); this method is
   common to all parsers */
void Parser::closeFile() {
   if (parseFile != NULL) {
      if (parseFile != stdin) {
         fclose(parseFile);
      }
      parseFile = NULL;
      strcpy(curLine, "");
      strcpy(error, "");
      parseFileSize = 0;
      parseFilePos = 0;
      parsedlines = 0;
      readlines = 0;
   }
}

/* Returns the number of lines that have been successfully parsed
   (e.g. contain useful data) until now; this method is common to all
   parsers */
unsigned long int Parser::parsedLines() {
   return parsedlines;
}

/* Returns the number of total lines that have been read from
   the file until now (including unuseful ones); this method is common
   to all parsers */
unsigned long int Parser::readLines() {
   return readlines;
}

/* If parseLine failed, this method returns a short explanation
   of the reason of the fault */
char *Parser::errReason() {
   return error;
};

// This method returns the line the parser is currently on (if any)
char *Parser::currentLine() {
   strcpy(copiedLine, curLine);
   return copiedLine;
};

// Returns current position in the file, in bytes
unsigned long int Parser::filePos() {
   return parseFilePos;
}

// Returns the size of the file being parsed, in bytes
unsigned long int Parser::fileSize() {
   return parseFileSize;
}

// ************** File parser ***************
