/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: tpathExtract.cc                              */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "tpathExtract.h"

/* Warning: using this code assumes that your C++ compiler supports
   partial expression evaluation (e.g. In (A || B) B is not evaluated
   at all if A is true */

/* This function parses each option's mode and sets its second argument
   accordingly */
bool parseMode (char *sourceString, unsigned char *mode) {
   char *curToken, string[strlen(sourceString + 1)];
   bool keep, fix, del, save, tag, set;
   keep = fix = del = save = tag = false;
   strcpy (string, sourceString);
   *mode = 0x0;
   curToken = strtok (string, ",");
   set = false;
   while (curToken != NULL) {
      set = false;
      if (strncmp(curToken, "keep", 4) == 0) {
         if (keep || tag || del || fix || save) {
            // Duplicate setting or incompatibility
            return false;
         }
         set = true;
         keep = true;
         *mode = 0x0;
      }
      if (strncmp(curToken, "fix", 6) == 0) {
         if (tag || del || keep || fix) {
            /* Incompatibility with delete, keep and tag
               or duplicate setting */
            return false;
         }
         set = true;
         fix = true;
         *mode |= 0x1;
      }
      if (strncmp(curToken, "delete", 6) == 0) {
         if (fix || tag || keep || del) {
            /* Incompatibility with fix, keep and tag
               or duplicate setting */
            return false;
         }
         set = true;
         del = true;
         *mode |= 0x2;
      }
      if (strncmp(curToken, "save", 4) == 0) {
         if (save || keep) {
            // Duplicate setting or incompatibility with keep
            return false;
         }
         set = true;
         save = true;
         *mode |= 0x4;
      }
      if (strncmp(curToken, "tag", 9) == 0) {
         if (fix || del || keep || tag) {
            /* Incompatibility with keep, fix and delete
               or duplicate setting */
            return false;
         }
         set = true;
         tag = true;
         *mode |= 0x8;
      }
      if (!set) {
         // Invalid mode string
         return false;
      }
      curToken = strtok (NULL, ",");
   }
   return set;
}

/* Very similar to the previous function, but with specific
   options for prefixes settings */
bool parsePfxMode (char *sourceString, extraction_mode *mode) {
   char *curToken, string[strlen(sourceString + 1)];
   bool set, show, unique, allmask, hide;
   strcpy (string, sourceString);
   (*mode).showPrefixes = false;
   (*mode).alwaysPrefix = false;
   (*mode).uniquePrefixes = false;
   curToken = strtok (string, ",");
   set = show = hide = unique = allmask = false;
   while (curToken != NULL) {
      set = false;
      if (strncmp(curToken, "show", 4) == 0) {
         if (show || hide) {
            // Incompatibility with hide or duplicate setting
            return false;
         }
         set = true;
         show = true;
         (*mode).showPrefixes = true;
      }
      if (strncmp(curToken, "unique", 6) == 0) {
         if (hide || unique) {
            // Incompatibility with hide or duplicate setting
            return false;
         }
         set = true;
         unique = true;
         (*mode).uniquePrefixes = true;
      }
      if (strncmp(curToken, "allmask", 7) == 0) {
         if (hide || allmask) {
            // Incompatibility with hide or duplicate setting
            return false;
         }
         set = true;
         allmask = true;
         (*mode).alwaysPrefix = true;
      }
      if (strncmp(curToken, "hide", 4) == 0) {
         if (unique || allmask || show || hide) {
            /* Incompatibility with show, unique or allmask
               or duplicate setting */
            return false;
         }
         set = true;
         hide = true;
         (*mode).showPrefixes = false;
      }
      if (!set) {
         // Invalid mode string
         return false;
      }
      curToken = strtok (NULL, ",");
   }
   if (!show && (unique || allmask)) {
      // Cannot use unique or allmask without show
      return false;
   }
   return set;
}

/* The following function parses input arguments and puts results into
   `mode'; if an error occurs, it returns false and fills `error' with
   a printable error message; otherwise, it returns true */
bool parseArguments (int argc, char **argv, extraction_mode *mode, char **error) {
   bool parsed, prepending, cycles, privateAS, sets,
        vantagePoint, forceSource, incomp, toFile, prefixes;
   long int value;
   char *err;
   // Set default parameters
   (*mode).prepending = 0x1;
   (*mode).cycles = 0x2;
   (*mode).privateAS = 0x1;
   (*mode).sets = 0x1;
   (*mode).prepAS = 0x0;
   (*mode).alwaysPrefix = 0x0;
   (*mode).incomplete = 0x0;
   (*mode).toFile = false;
   (*mode).showHelp = false;
   (*mode).verbose = false;
   (*mode).overwrite = false;
   (*mode).groupASpaths = false;
   (*mode).showPrefixes = false;
   (*mode).alwaysPrefix = false;
   (*mode).uniquePrefixes = false;
   forceSource = prepending = cycles = privateAS = sets = incomp =
      vantagePoint = toFile = prefixes = false;
   strcpy (*error, "");
   for (int i = 1; i <= argc - 1; i++) {
      parsed = false;
      if (forceSource) {
         if ((*mode).sourceFile != NULL) {
            strcat (*error, "There is more than one input file");
            return false;
         }
         (*mode).sourceFile = new char[strlen(argv[i]) + 1];
         strcpy ((*mode).sourceFile, argv[i]);
         parsed = true;
         forceSource = false;
      }
      if (toFile) {
         if ((*mode).outFile != NULL) {
            strcat (*error, "There is more than one output file");
            return false;
         }
         (*mode).outFile = new char[strlen(argv[i]) + 1];
         strcpy ((*mode).outFile, argv[i]);
         parsed = true;
         toFile = false;
      }
      if ((*mode).prepASnum == NULL && ((*mode).prepAS != 0x0)) {
         // Vantage point number expected
         (*mode).prepASnum = new unsigned short int;
         value = strtol(argv[i], &err, 10);
         if (value < 0 || value > 65535 || (strlen(err) > 0)) {
            strcat (*error, "Invalid number for vantage point: `");
            strcat (*error, argv[i]);
            strcat (*error, "'");
            return false;
         }
         *((*mode).prepASnum) = value;
         parsed = true;
      }
      if ((*mode).prepFile == NULL && ((*mode).prepending & 0x4)) {
         // Prepending file name expected
         (*mode).prepFile = new char[strlen(argv[i]) + 1];
         strcpy ((*mode).prepFile, argv[i]);
         parsed = true;
      }
      if ((*mode).cycleFile == NULL && ((*mode).cycles & 0x4)) {
         // Cycle file name expected
         (*mode).cycleFile = new char[strlen(argv[i]) + 1];
         strcpy ((*mode).cycleFile, argv[i]);
         parsed = true;
      }
      if ((*mode).privFile == NULL && ((*mode).privateAS & 0x4)) {
         // Private file name expected
         (*mode).privFile = new char[strlen(argv[i]) + 1];
         strcpy ((*mode).privFile, argv[i]);
         parsed = true;
      }
      if ((*mode).setFile == NULL && ((*mode).sets & 0x4)) {
         // Sets file name expected
         (*mode).setFile = new char[strlen(argv[i]) + 1];
         strcpy ((*mode).setFile, argv[i]);
         parsed = true;
      }
      if ((*mode).incompFile == NULL && ((*mode).incomplete & 0x4)) {
         // Incomplete file name expected
         (*mode).incompFile = new char[strlen(argv[i]) + 1];
         strcpy ((*mode).incompFile, argv[i]);
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "--help") == 0) {
         (*mode).showHelp = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-o") == 0) {
         toFile = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-f") == 0) {
         if ((*mode).overwrite) {
            strcat (*error, "Option -f is used at least twice");
            return false;
         }
         (*mode).overwrite = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-v") == 0) {
         if ((*mode).verbose) {
            strcat (*error, "Option -v is used at least twice");
            return false;
         }
         (*mode).verbose = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "-s") == 0) {
         forceSource = true;
         parsed = true;
      }
      if (!parsed && strcmp(argv[i], "--group") == 0) {
         if ((*mode).groupASpaths) {
            strcat (*error, "Option --group is used at least twice");
            return false;
         }
         (*mode).groupASpaths = true;
         parsed = true;
      }
      if (!parsed && strncmp(argv[i], "--prefixes", 10) == 0) {
         if (argv[i][10] != '=') {
            strcat (*error, "`=' expected after --prefixes option");
            return false;
         }
         parsed = parsePfxMode(argv[i]+11, mode);
         if (!parsed) {
            strcat (*error, "Invalid mode for --prefixes: `");
            strcat (*error, argv[i]+11);
            strcat (*error, "'");
            return false;
         }
         if (prefixes) {
            strcat (*error, "Repeated use of --prefixes option");
            return false;
         }
         prefixes = true;
         parsed = true;
      }
      if (!parsed && strncmp(argv[i], "--prepending", 12) == 0) {
         if (argv[i][12] != '=') {
            strcat (*error, "`=' expected after --prepending option");
            return false;
         }
         parsed = parseMode(argv[i]+13, &((*mode).prepending));
         if (!parsed) {
            strcat (*error, "Invalid mode for --prepending: `");
            strcat (*error, argv[i]+13);
            strcat (*error, "'");
            return false;
         }
         if (prepending) {
            strcat (*error, "Repeated use of --prepending option");
            return false;
         }
         prepending = true;
      }
      if (!parsed && strncmp(argv[i], "--cycles", 8) == 0) {
         if (argv[i][8] != '=') {
            strcat (*error, "`=' expected after --cycles option");
            return false;
         }
         parsed = parseMode(argv[i]+9, &((*mode).cycles));
         if (!parsed) {
            strcat (*error, "Invalid mode for --cycles: `");
            strcat (*error, argv[i]+9);
            strcat (*error, "'");
            return false;
         }
         if (cycles) {
            strcat (*error, "Repeated use of --cycles option");
            return false;
         }
         cycles = true;
      }
      if (!parsed && strncmp(argv[i], "--privas", 8) == 0) {
         if (argv[i][8] != '=') {
            strcat (*error, "`=' expected after --privas option");
            return false;
         }
         parsed = parseMode(argv[i]+9, &((*mode).privateAS));
         if (!parsed) {
            strcat (*error, "Invalid mode for --privas: `");
            strcat (*error, argv[i]+9);
            strcat (*error, "'");
            return false;
         }
         if (privateAS) {
            strcat (*error, "Repeated use of --privas option");
            return false;
         }
         privateAS = true;
      }
      if (!parsed && strncmp(argv[i], "--sets", 6) == 0) {
         if (argv[i][6] != '=') {
            strcat (*error, "`=' expected after --sets option");
            return false;
         }
         parsed = parseMode(argv[i]+7, &((*mode).sets));
         if (!parsed) {
            strcat (*error, "Invalid mode for --sets: `");
            strcat (*error, argv[i]+7);
            strcat (*error, "'");
            return false;
         }
         if (sets) {
            strcat (*error, "Repeated use of --sets option");
            return false;
         }
         if ((*mode).sets == 0x0) {
            strcat (*error, "AS sets cannot be kept");
            return false;
         }
         sets = true;
      }
      if (!parsed && strncmp(argv[i], "--incomplete", 12) == 0) {
         if (argv[i][12] != '=') {
            strcat (*error, "`=' expected after --incomplete option");
            return false;
         }
         parsed = parseMode(argv[i]+13, &((*mode).incomplete));
         if (!parsed) {
            strcat (*error, "Invalid mode for --incomplete: `");
            strcat (*error, argv[i]+13);
            strcat (*error, "'");
            return false;
         }
         if (incomp) {
            strcat (*error, "Repeated use of --sets option");
            return false;
         }
         if ((*mode).incomplete & 0x1) {
            strcat (*error, "Incompleteness cannot be fixed");
            return false;
         }
         incomp = true;
      }
      if (!parsed && strncmp(argv[i], "--vpoint", 8) == 0) {
         if (argv[i][8] != '=') {
            strcat (*error, "`=' expected after --vpoint option");
            return false;
         }
         if (strncmp(argv[i] + 9, "none", 4) == 0) {
            (*mode).prepAS = 0x0;
            if ((*mode).prepASnum != NULL) {
               delete (*mode).prepASnum;
               (*mode).prepASnum = NULL;
            }
         }
         else {
            if (strncmp(argv[i] + 9, "always", 5) == 0) {
               (*mode).prepAS = 0x1;
            }
            else {
               if (strncmp(argv[i] + 9, "auto", 4) == 0) {
                  (*mode).prepAS = 0x2;
               }
               else {
                  strcat (*error, "Invalid mode for --vpoint: `");
                  strcat (*error, argv[i] + 9);
                  strcat (*error, "'");
                  return false;
               }
            }
         }
         if (vantagePoint) {
            strcat (*error, "Repeated use of --vpoint option");
            return false;
         }
         parsed = true;
         vantagePoint = true;
      }

      if (!parsed) {
         // Perhaps this is input file name
         if ((*mode).sourceFile != NULL) {
            strcat (*error, "Unrecognized option: `");
            strcat (*error, argv[i]);
            strcat (*error, "'");
            return false;
         }
         (*mode).sourceFile = new char[strlen(argv[i]) + 1];
         strcpy ((*mode).sourceFile, argv[i]);
      }
   }
   if (!(*mode).groupASpaths && (*mode).uniquePrefixes) {
      strcat (*error, "Use of `unique' flag is senseless without --group");
      return false;
   }
   if ((*mode).sets == 0x4) {
      strcat (*error, "Cannot just save paths with AS sets to file");
      return false;
   }
   if ((*mode).prepFile == NULL && ((*mode).prepending & 0x4)) {
      strcat (*error, "I need a name for the file to save paths with prepending in");
      return false;
   }
   if ((*mode).cycleFile == NULL && ((*mode).cycles & 0x4)) {
      strcat (*error, "I need a name for the file to save cycling paths in");
      return false;
   }
   if ((*mode).privFile == NULL && ((*mode).privateAS & 0x4)) {
      strcat (*error, "I need a name for the file to save paths with private ASs in");
      return false;
   }
   if ((*mode).setFile == NULL && ((*mode).sets & 0x4)) {
      strcat (*error, "I need a name for the file to save paths with AS sets in");
      return false;
   }
   if ((*mode).incompFile == NULL && ((*mode).incomplete & 0x4)) {
      strcat (*error, "I need a name for the file to save incomplete paths in");
      return false;
   }
   if ((*mode).prepASnum == NULL && ((*mode).prepAS != 0x0)) {
      strcat (*error, "Vantage point number expected");
      return false;
   }
   return true;
}

// This function just prints a small header
void printHeader (FILE *file) {
   // The following avoids printing header another time
   static bool alreadyPrinted = false;
   if (!alreadyPrinted) {
      if (isTerm(file)) {
         fprintf (file, "%stpathExtract%s 1.0\n", CONSOLE_BOLD, CONSOLE_NORMAL);
      }
      else {
         fprintf (file, "tpathExtract 1.0\n");
      }
   }
   alreadyPrinted = true;
}

// This functions show instructions about program usage
void printHelp (FILE *file, char *cmdLine) {
   char CON_BOLD[strlen(CONSOLE_BOLD) + 1];
   char CON_UNDERLINE[strlen(CONSOLE_UNDERLINE) + 1];
   char CON_NORMAL[strlen(CONSOLE_NORMAL) + 1];
   printHeader (file);
   if (isTerm(file)) {
      strcpy (CON_BOLD, CONSOLE_BOLD);
      strcpy (CON_UNDERLINE, CONSOLE_UNDERLINE);
      strcpy (CON_NORMAL, CONSOLE_NORMAL);
   }
   else {
      strcpy (CON_BOLD, "");
      strcpy (CON_UNDERLINE, "");
      strcpy (CON_NORMAL, "");
   }
   fprintf (file, "%sTORQUE toolkit%s AS path extractor\n", CON_BOLD, CON_NORMAL);
   fprintf (file, "USAGE: %s [settings] [options] [[-s] <filename>]\n\n", cmdLine);
   fprintf (file, "%sSettings%s are:\n", CON_UNDERLINE, CON_NORMAL);
   fprintf (file, "  --group\n");
   fprintf (file, "      By default, output is a list of AS paths, including repetitions;\n");
   fprintf (file, "      using --group option reports each AS path once, together with the\n");
   fprintf (file, "      number of its occurrences in the BGP dump. When requested, the path\n");
   fprintf (file, "      is tagged with a merging of the tags associated with each instance\n");
   fprintf (file, "      of its. This means that the number of occurrences is just referred\n");
   fprintf (file, "      to the path itself, while the presence of a tag means that at least\n");
   fprintf (file, "      one of the occurrences had that property. If you wish to exactly\n");
   fprintf (file, "      know how many paths have that tag, either use -v option or save\n");
   fprintf (file, "      paths to external files\n\n");
   fprintf (file, "  --prefixes={show,allmask,unique,hide}\n");
   fprintf (file, "      show:    turns on prefix logging (i.e., for each path, a list of the\n");
   fprintf (file, "               prefixes that followed that path is reported); if --group is\n");
   fprintf (file, "               not used, a single prefix is reported for each AS path\n");
   fprintf (file, "      allmask: append a netmask to every network (even those without a\n");
   fprintf (file, "               netmask)\n");
   fprintf (file, "      unique:  report every prefix at most one time for each path\n");
   fprintf (file, "               (warning: a prefix might be reported more than one time\n");
   fprintf (file, "               even when using this option if it appears both with and\n");
   fprintf (file, "               without netmask in the source file); this mode should be\n");
   fprintf (file, "               used only in combination with --group\n");
   fprintf (file, "      Don't use:  show    with    hide\n");
   fprintf (file, "                  allmask without show\n");
   fprintf (file, "                  unique  without show\n\n");
   fprintf (file, "  --vpoint=always|auto|none\n");
   fprintf (file, "      none:   don't change AS paths\n");
   fprintf (file, "      always: prepend a user provided AS number to each AS path before\n");
   fprintf (file, "              processing it\n");
   fprintf (file, "      auto:   only prepend user provided AS if the first AS in the path is\n");
   fprintf (file, "              different\n");
   fprintf (file, "  When using either `always' or `auto', the following argument must be a\n");
   fprintf (file, "  valid AS number.\n\n");
   fprintf (file, "  The following settings all use the same modes:\n");
   fprintf (file, "      keep:   leave path unchanged\n");
   fprintf (file, "      fix:    remove feature from the path, but do not remove path\n");
   fprintf (file, "      delete: remove path from main output\n");
   fprintf (file, "      save:   also save (unchanged) path to file; when using this mode, the\n");
   fprintf (file, "              following parameter must be a file name\n");
   fprintf (file, "      tag:    tag the path using a letter\n");
   fprintf (file, "      Don't use:  keep   with any combination of the other modes\n");
   fprintf (file, "                  fix    with delete or tag\n");
   fprintf (file, "                  delete with tag\n\n");
   fprintf (file, "  --incomplete={keep,delete,save,tag}\n");
   fprintf (file, "      Perform listed operations on paths marked as incomplete; if\n");
   fprintf (file, "      requested, paths are tagged with an 'I'\n\n");
   fprintf (file, "  --prepending={keep,fix,delete,save,tag}\n");
   fprintf (file, "      Perform listed operations on paths with prepending; using `fix' also\n");
   fprintf (file, "      deletes \"false\" prepending from AS sets; if requested, paths are\n");
   fprintf (file, "      tagged with a 'P'\n\n");
   fprintf (file, "  --privas={keep,fix,delete,save,tag}\n");
   fprintf (file, "      Perform listed operations on paths with private ASs; use of `fix'\n");
   fprintf (file, "      implies removal of private ASs; if requested, paths are tagged with\n");
   fprintf (file, "      a 'p'\n\n");
   fprintf (file, "  --sets={fix,delete,save,tag}\n");
   fprintf (file, "      Perform listed operations on paths with AS sets; `fix' deletes the\n");
   fprintf (file, "      whole set from the path; `save' cannot be used without either\n");
   fprintf (file, "      combination of the other modes; if requested, paths are tagged\n");
   fprintf (file, "      with a 'S'\n\n");
   fprintf (file, "  --cycles={keep,fix,delete,save,tag}\n");
   fprintf (file, "      Perform listed operations on paths with cycles; cycle checking is\n");
   fprintf (file, "      applied after all the other changes took place, and completely\n");
   fprintf (file, "      skips ASs in AS sets; using `fix' deletes every instance of an\n");
   fprintf (file, "      AS in the path but the first; if requested, paths are tagged with\n");
   fprintf (file, "      a 'C'\n\n");
   fprintf (file, "Note: modes included in {} can be combined together, using comma as\n");
   fprintf (file, "separator; incompatible combinations are listed above; when using any of\n");
   fprintf (file, "the above options, corresponding built-in defaults are automatically\n");
   fprintf (file, "overridden (e.g., that is why you cannot use `save' and `keep' together)\n\n\n");
   fprintf (file, "%sOptions%s are:\n", CON_UNDERLINE, CON_NORMAL);
   fprintf (file, "  -f      Don't ask user confirmation (default behaviour) when overwriting\n");
   fprintf (file, "          existing files\n");
   fprintf (file, "  -o <filename>  Save main output to file instead of printing it on stdout\n");
   fprintf (file, "  -v      Verbose mode (shows both parsing errors and final statistics);\n");
   fprintf (file, "          default mode is non-verbose\n");
   fprintf (file, "  --help  Show this help\n\n\n");
   fprintf (file, "<filename> is the name of a BGP dump file. Omitting it results in input\n");
   fprintf (file, "being taken from stdin.\n");
   fprintf (file, "Using of -s is suggested when filename may be confused with other command\n");
   fprintf (file, "line options.\n\n");
   fprintf (file, "Default behaviour corresponds to the use of the following settings/options:\n");
   fprintf (file, "--prefixes=hide --incomplete=keep --prepending=fix --privas=fix\n");
   fprintf (file, "--sets=fix --cycles=delete\n");
}

// Print a prefix, with network and (if either forced or needed) netmask
void printPrefix (FILE *file, IPAddress addr, IPNetmask *mask, bool alwaysPrefix) {
   IPPrefix pref;
   unsigned char *bytes;
   bytes = addr.getAddress();
   if (alwaysPrefix) {
      if (mask != NULL) {
         fprintf (file, "%i.%i.%i.%i/%i ",
            bytes[0], bytes[1], bytes[2], bytes[3],
            mask->brief());
      }
      else {
         pref = IPPrefix(addr);
         fprintf (file, "%i.%i.%i.%i/%i ",
            bytes[0], bytes[1], bytes[2], bytes[3],
            pref.netmask().brief());
      }
   }
   else {
      if (mask != NULL) {
         fprintf (file, "%i.%i.%i.%i/%i ",
            bytes[0], bytes[1], bytes[2], bytes[3],
            mask->brief());
      }
      else {
         fprintf (file, "%i.%i.%i.%i ",
            bytes[0], bytes[1], bytes[2], bytes[3]);
      }
   }
}

// This function prints an AS path
void printASpath (FILE *file, List<bgpParser::ASItem> *path) {
   bool goOn, inSet;
   goOn = !path->isEmpty();
   inSet = false;
   for (path->goToFirst(); goOn; goOn = path->next()) {
      if (!path->isFirst()) {
         fprintf (file, " ");
      }
      if (!inSet && (path->current().flags & AS_SET)) {
         // Start AS set
         inSet = true;
         fprintf (file, "(");
      }
      if (inSet && !(path->current().flags & AS_SET)) {
         // End AS set
         inSet = false;
         fprintf (file, ")");
      }
      fprintf (file, "%i", path->current().ASnum);
   }
   if (inSet) {
      fprintf (file, ")");
   }
}

// This function prepares an output file for writing into it
int openWriteFile(FILE **returnedStream, char *fileName, bool alwaysOverwrite) {
   bool writeFile;
   FILE *file;
   if (fileName != NULL) {
      // Test if file exists
      writeFile = true;
      file = fopen (fileName, "r");
      if (file != NULL) {
         fclose (file);
         if (!alwaysOverwrite) {
            printHeader (stderr);
            writeFile = askOverwrite (fileName);
         }
      }
      if (writeFile) {
         (*returnedStream) = fopen (fileName, "w");
         if ((*returnedStream) == NULL) {
            printHeader (stderr);
            fprintf (stderr, "Error: Unable to open file `%s' for writing\n", fileName);
            fprintf (stderr, "(%s)\n", strerror(errno));
            return 2;
         }
      }
      else {
         return 3;
      }
   }
   return 0;
}

/* This function takes as input two AS paths and returns true if
   they are the same path (including the tags) */
bool samePath (List<bgpParser::ASItem> &p1, List<bgpParser::ASItem> &p2) {
   bool goOn;
   if (p1.len() != p2.len()) {
      return false;
   }
   p1.goToFirst();
   p2.goToFirst();
   goOn = !p1.isEmpty() && !p2.isEmpty();
   for (; goOn; goOn = p1.next() && p2.next()) {
      if (p1.current().ASnum != p2.current().ASnum ||
          p1.current().flags != p2.current().flags) {
         return false;
      }
   }
   return true;
}

// This function updates a map with the current AS path
void updateMap (BST<List<pathData*>*, unsigned short int> &pathSet,
                List<bgpParser::ASItem> &ASpath, char *tagStr, IPAddress network,
                IPNetmask *netmask, bool uniquePrefixes) {
   List<pathData*> *newDataList, **curDataList;
   pathData *curPath;
   IPNetmask *newMask, *lastMask;
   bool goOn, done;
   curDataList = pathSet.findItem(ASpath.last().ASnum);
   if (curDataList == NULL) {
      newDataList = new List<pathData*>;
      curPath = new pathData;
      curPath->ASpath = ASpath;
      curPath->howMany = 1;
      curPath->tags = new char[7];
      strcpy (curPath->tags, tagStr);
      curPath->network.enqueue(network);
      if (netmask != NULL) {
         // Current prefix also has netmask
         newMask = new IPNetmask(*netmask);
         curPath->netmask.enqueue(newMask);
      }
      else {
         // Current prefix has no netmask
         curPath->netmask.enqueue((IPNetmask*) NULL);
      }
      newDataList->enqueue(curPath);
      pathSet.addItem(newDataList, ASpath.last().ASnum);
   }
   else {
      goOn = true;
      done = false;
      for ((*curDataList)->goToFirst(); goOn && !done; goOn = (*curDataList)->next()) {
         curPath = (*curDataList)->current();
         if (samePath(curPath->ASpath, ASpath)) {
            // Path already exists in the map
            done = true;
            curPath->howMany++;
            // Merge current tag string with existing one
            for (int i=0; i<=6; i++) {
               if (tagStr[i]!=' ') {
                  curPath->tags[i] = tagStr[i];
               }
            }

            if (!uniquePrefixes || curPath->network.last() != network) {
               /* Either user did not ask for prefixes uniqueness, or
                  network differs from last inserted prefix */
               curPath->network.enqueue(network);
               if (netmask != NULL) {
                  newMask = new IPNetmask(*netmask);
                  curPath->netmask.enqueue(newMask);
               }
               else {
                  curPath->netmask.enqueue((IPNetmask*) NULL);
               }
            }
            else {
               // Same network, and user asked for prefix uniqueness
               lastMask = curPath->netmask.last();
               if (((netmask != NULL) && (lastMask != NULL) && (*netmask != *lastMask)) ||
                   ((netmask != NULL) && (lastMask == NULL)) ||
                   ((netmask == NULL) && (lastMask != NULL))) {
                  curPath->network.enqueue(network);
                  if (netmask != NULL) {
                     newMask = new IPNetmask(*netmask);
                     curPath->netmask.enqueue(newMask);
                  }
                  else {
                     curPath->netmask.enqueue((IPNetmask*) NULL);
                  }
               }
            }
         }
      }
      if (!done) {
         /* Map already contains a bin for ASs with the same
            last AS as the current one, but does not
            contain current AS */
         curPath = new pathData;
         curPath->ASpath = ASpath;
         curPath->howMany = 1;
         curPath->tags = new char[7];
         strcpy (curPath->tags, tagStr);
         curPath->network.enqueue(network);
         if (netmask != NULL) {
            // Current prefix also has netmask
            newMask = new IPNetmask(*netmask);
            curPath->netmask.enqueue(newMask);
         }
         else {
            // Current prefix has no netmask
            curPath->netmask.enqueue((IPNetmask*) NULL);
         }
         (*curDataList)->enqueue(curPath);
      }
   }
}

/* This function either writes to file or updates the
   corresponding path map (depending on the use of
   the --group option) */
void writeToFile (FILE *file, char *tagStr, bool groupASpaths,
                  bool showPrefixes, bool alwaysPrefix,
                  bool uniquePrefixes, bool tag,
                  IPAddress network, IPNetmask *netmask,
                  List<bgpParser::ASItem> &originalPath,
                  BST<List<pathData*>*, unsigned short int> &pathSet) {
   if (!groupASpaths) {
      if (tag) {
         fprintf (file, "%s", tagStr);
      }
      printASpath (file, &originalPath);
      if (showPrefixes) {
         fprintf (file, "\t");
         printPrefix (file, network, netmask, alwaysPrefix);
      }
      fprintf (file, "\n");
   }
   else {
      updateMap (pathSet, originalPath, tagStr, network, netmask, uniquePrefixes);
   }
}

/* This function prints the contents of a path map to a file
   and at the same time releases memory that had been allocated
   for it */
void printMap (FILE *file, BST<List<pathData*>*, unsigned short int> &pathSet,
               bool tag, bool showPercentage, bool showPrefixes, bool alwaysPrefix) {
   List<List<pathData*>*> pathDataPointerList;
   List<pathData*> *pathDataList;
   pathData *curPathData;
   IPAddress addr;
   IPNetmask *mask;
   unsigned int n, originalLen;
   pathDataPointerList = pathSet.listItems();
   pathDataPointerList.goToFirst();
   n = 0;
   originalLen = pathDataPointerList.len();
   while (!pathDataPointerList.isEmpty()) {
      n++;
      if (showPercentage) {
         // Print percentage of completion
         printPercentage((float) n, (float) originalLen);
      }
      pathDataList = pathDataPointerList.pop();
      pathDataList->goToFirst();
      while (!pathDataList->isEmpty()) {
         curPathData = pathDataList->pop();
         fprintf (file, "%lu\t", curPathData->howMany);
         if (tag) {
            fprintf (file, "%s\t", curPathData->tags);
         }
         delete curPathData->tags;
         printASpath (file, &(curPathData->ASpath));
         if (showPrefixes) {
            fprintf (file, "\t");
         }
         while (!(curPathData->network).isEmpty()) {
            addr = (curPathData->network).pop();
            mask = (curPathData->netmask).pop();
            if (showPrefixes) {
               printPrefix (file, addr, mask, alwaysPrefix);
            }
            if (mask != NULL) {
               delete mask;
            }
         }
         fprintf (file, "\n");
         delete curPathData;
      }
      delete pathDataList;
   }
}



int main(int argc, char **argv) {
   extraction_mode params;
   bool goOn, done, log, tag, stdinIsTerm, removeIt,
        stdoutIsTerm, keepPath, showPercentage,
        tagCycle, tagPrep, tagPriv, tagSet, tagIncomp,
        writeCycle, writePrep, writePriv, writeSet, writeIncomp;
   FILE *cycleFile, *prepFile, *privFile, *setFile, *incompFile, *outFile;
   bgpParser parser;
   int i;
   /* The following variables are used to store statistics when
      in verbose mode */
   unsigned long int rep[6][8], wrongLines, correctLines;
   bool increase[6][8];
   int prepLevel, privLevel, setLevel, cycleLevel;

   char *err = new char [256];
   char tagStr[7], oldTagStr[7];
   unsigned short int lastAS;

   /* These declarations are used when grouping AS paths, in order
      to map and quickly retrieve them */
   BST<List<pathData*>*, unsigned short int> pathSet[6];

   /* In the following declaration, char is used to minimize
      memory waste (key itself is sufficient to store
      AS number) */
   BST<char, unsigned short int> pathASs;
   List<bgpParser::ASItem> originalPath;
   bgpParser::ASItem curAS;

   if (!parseArguments (argc, argv, &params, &err)) {
      printHeader(stderr);
      fprintf (stderr, "Syntax error: %s\n", err);
      fprintf (stderr, "Type `%s --help' for more information\n", argv[0]);
      delete err;
      return 1;
   }
   delete [] err;

   if (params.showHelp) {
      printHeader (stdout);
      printHelp (stdout, argv[0]);
      return 0;
   }

   /* Now determine wheter stdin/stdout are connected to a terminal
      or to a file */
   stdinIsTerm = isTerm(stdin);
   stdoutIsTerm = isTerm(stdout);

   // Only report status messages when stdout is connected to a terminal
   log = stdoutIsTerm;

   // Now open all the needed files
   if (params.sourceFile == NULL) {
      // Get input from standard input, whatever it is
      parser.openStdin();
   }
   else {
      if (!parser.openFile(params.sourceFile)) {
         printHeader (stderr);
         fprintf (stderr, "Error: Unable to open input file: `%s'\n", params.sourceFile);
         fprintf (stderr, "(%s)\n", strerror(errno));
         return 2;
      }
   }

   if (params.outFile != NULL) {
      i = openWriteFile (&outFile, params.outFile, params.overwrite);
   }
   else {
      outFile = stdout;
      i = 0;
   }
   if (i != 0) {
      return i;
   }

   i = openWriteFile (&prepFile, params.prepFile, params.overwrite);
   if (i != 0) {
      return i;
   }

   i = openWriteFile (&cycleFile, params.cycleFile, params.overwrite);
   if (i != 0) {
      return i;
   }

   i = openWriteFile (&privFile, params.privFile, params.overwrite);
   if (i != 0) {
      return i;
   }

   i = openWriteFile (&setFile, params.setFile, params.overwrite);
   if (i != 0) {
      return i;
   }

   i = openWriteFile (&incompFile, params.incompFile, params.overwrite);
   if (i != 0) {
      return i;
   }

   if (log) {
      printHeader (stdout);
   }

   for (int x = 0; x <= 5; x++) {
      for (int y = 0; y <= 7; y++) {
         rep[x][y] = 0;
         increase[x][y] = false;
      }
   }
   correctLines = wrongLines = 0;
   tag = params.prepending & 0x8 ||
         params.cycles & 0x8 ||
         params.privateAS & 0x8 ||
         params.sets & 0x8 ||
         params.incomplete & 0x8;
   showPercentage = stdoutIsTerm && stdinIsTerm && params.outFile != NULL;

   if (showPercentage) {
      if (params.groupASpaths) {
         printf ("Reading input file...\n");
      }
      else {
         printf ("Processing...\n");
      }
   }
   while (parser.nextLine() == PARSEFILE_NOERROR) {
      if (showPercentage) {
         // Print percentage of completion
         printPercentage((float) parser.filePos(), (float) parser.fileSize());
      }

      if (parser.parseLine() != PARSE_NOERROR) {
         wrongLines++;
         if (params.verbose) {
            fprintf (stderr, "\n");
            printHeader (stderr);
            fprintf (stderr, "Parse error in line %lu of file `%s':\n",
               parser.readLines(), params.sourceFile);
            fprintf (stderr, "%s\n", parser.currentLine());
            fprintf (stderr, "%s\n", parser.errReason());
         }
      }
      else {
         correctLines++;
         keepPath = true;
         /* Empty the set of ASs in the path (this is used to
            check for cycles */
         pathASs = BST<char, unsigned short int>();
         writeCycle = writePrep = writePriv = writeSet = writeIncomp = false;
         tagCycle = tagPrep = tagPriv = tagSet = tagIncomp = false;
         strcpy (tagStr, "");
         strcpy (oldTagStr, "");

         // Is the path incomplete?
         if (parser.parsedData.path_flags & ORIGIN_INCOMPLETE) {
            // Path is marked as incomplete
            keepPath = !(params.incomplete & 0x2);
            if (keepPath) {
               rep[0][0]++;
            }
            else {
               rep[0][2]++;
            }
            writeIncomp = (params.incomplete & 0x4);
            if (writeIncomp) {
               rep[0][3]++;
            }
            if (params.incomplete & 0x8) {
               // Flag the path
               rep[0][4]++;
               tagIncomp = true;
            }
         }

         // Prepend user provided AS
         if (((params.prepAS & 0x2) && (parser.parsedData.ASpath.isEmpty() ||
             (parser.parsedData.ASpath.first().ASnum != *(params.prepASnum)))) ||
            (params.prepAS & 0x1)) {
            curAS.ASnum = *(params.prepASnum);
            curAS.flags = 0x0;
            if (curAS.ASnum >= 64512) {
               // User provided AS is a private one
               curAS.flags |= AS_PRIVATE;
            }
            parser.parsedData.ASpath.push(curAS);
         }

         originalPath = parser.parsedData.ASpath;
         lastAS = originalPath.first().ASnum;

         // Now examine parsed AS path
         parser.parsedData.ASpath.goToFirst();
         goOn = !parser.parsedData.ASpath.isEmpty();
         /* The following variables are used in order to
            keep trace of the real properties of the path
            (after processing) */
         prepLevel = privLevel = setLevel = cycleLevel = 0;
         while (goOn) {
            curAS = parser.parsedData.ASpath.current();
            done = false;
            removeIt = false;
            if (curAS.flags & AS_SET) {
               // There is an AS set
               setLevel++;
               if (params.sets & 0x1) {
                  // Remove it
                  increase[3][1] = true;
                  setLevel--;
                  removeIt = true;
               }
               if (params.sets & 0x2) {
                  // Delete path (from main output)
                  increase[3][2] = true;
                  keepPath = false;
               }
               if (params.sets & 0x4) {
                  // Save path to file (with private ASs)
                  increase[3][3] = true;
                  writeSet = true;
               }
               if (params.sets & 0x8) {
                  // Flag the path
                  increase[3][4] = true;
                  tagSet = true;
               }
            }
            if (curAS.ASnum == lastAS && !parser.parsedData.ASpath.isFirst()) {
               // There is prepending
               prepLevel++;
               if (params.prepending == 0x0) {
                  increase[1][0] = true;
               }
               if (params.prepending & 0x1) {
                  // Remove it
                  increase[1][1] = true;
                  prepLevel--;
                  removeIt = true;
               }
               if (params.prepending & 0x2) {
                  // Delete path (from main output)
                  increase[1][2] = true;
                  keepPath = false;
               }
               if (params.prepending & 0x4) {
                  // Save path to file (with prepending)
                  increase[1][3] = true;
                  writePrep = true;
               }
               if (params.prepending & 0x8) {
                  // Flag the path
                  increase[1][4] = true;
                  tagPrep = true;
               }
            }
            if (curAS.flags & AS_PRIVATE) {
               // There is a private AS
               privLevel++;
               if (params.privateAS == 0x0) {
                  increase[2][0] = true;
               }
               if (params.privateAS & 0x1) {
                  // Remove it
                  increase[2][1] = true;
                  privLevel--;
                  removeIt = true;
               }
               if (params.privateAS & 0x2) {
                  // Delete path (from main output)
                  increase[2][2] = true;
                  keepPath = false;
               }
               if (params.privateAS & 0x4) {
                  // Save path to file (with private ASs)
                  increase[2][3] = true;
                  writePriv = true;
               }
               if (params.privateAS & 0x8) {
                  // Flag the path
                  increase[2][4] = true;
                  tagPriv = true;
               }
            }
            if (removeIt) {
               goOn = !parser.parsedData.ASpath.isLast();
               parser.parsedData.ASpath.removeCurrent();
               done = true;
            }
            else {
               // If no other change took place, check for cycles
               if (!(curAS.flags & AS_SET) && (curAS.ASnum != lastAS) &&
                   !pathASs.addItem(0, curAS.ASnum)) {
                  // There is a cycle
                  cycleLevel++;
                  if (params.cycles == 0x0) {
                     increase[4][0] = true;
                  }
                  if (params.cycles & 0x1) {
                     // Remove it
                     increase[4][1] = true;
                     cycleLevel--;
                     goOn = !parser.parsedData.ASpath.isLast();
                     parser.parsedData.ASpath.removeCurrent();
                     done = true;
                  }
                  if (params.cycles & 0x2) {
                     // Delete path (from main output)
                     increase[4][2] = true;
                     keepPath = false;
                  }
                  if (params.cycles & 0x4) {
                     // Save path to file (with cycles)
                     increase[4][3] = true;
                     writeCycle = true;
                  }
                  if (params.cycles & 0x8) {
                     // Flag the path
                     increase[4][4] = true;
                     tagCycle = true;
                  }
               }
            }
            if (!done) {
               lastAS = parser.parsedData.ASpath.current().ASnum;
               goOn = parser.parsedData.ASpath.next();
            }
         }

         // Update statistics
         for (int x = 1; x <= 4; x++) {
            for (int y = 0; y <= 4; y++) {
               if (increase[x][y]) {
                  increase[x][y] = false;
                  rep[x][y]++;
               }
            }
         }

         // Build tag string for both processed and original paths
         strcat (oldTagStr, (tagCycle ? "C" : " "));
         tagCycle &= (cycleLevel > 0);
         strcat (tagStr, (tagCycle ? "C" : " "));

         strcat (oldTagStr, (tagPrep ? "P" : " "));
         tagPrep &= (prepLevel > 0);
         strcat (tagStr, (tagPrep ? "P" : " "));

         strcat (oldTagStr, (tagPriv ? "p" : " "));
         tagPriv &= (privLevel > 0);
         strcat (tagStr, (tagPriv ? "p" : " "));
         
         strcat (oldTagStr, (tagSet ? "S" : " "));
         tagSet &= (setLevel > 0);
         strcat (tagStr, (tagSet ? "S" : " "));
         
         strcat (oldTagStr, (tagIncomp ? "I" : " "));
         strcat (tagStr, (tagIncomp ? "I" : " "));
         
         strcat (oldTagStr, " ");
         strcat (tagStr, " ");

         // Test if path is not empty
         if (!parser.parsedData.ASpath.isEmpty()) {
            if (keepPath) {
               rep[5][5]++;
               if (!params.groupASpaths) {
                  if (tag) {
                     fprintf (outFile, "%s", tagStr);
                  }
                  printASpath (outFile, &(parser.parsedData.ASpath));
                  if (params.showPrefixes) {
                     fprintf (outFile, "\t");
                     printPrefix (outFile, parser.parsedData.network,
                                  parser.parsedData.netmask, params.alwaysPrefix);
                  }
                  fprintf (outFile, "\n");
               }
               else {
                  updateMap (pathSet[0], parser.parsedData.ASpath, tagStr,
                             parser.parsedData.network, parser.parsedData.netmask,
                             params.uniquePrefixes);
               }
            }
         }
         else {
            // Just keep trace of discarded empty paths
            if (keepPath) {
               rep[5][7]++;
            }
            if (writeIncomp) {
               rep[0][7]++;
            }
            if (writePrep) {
               rep[1][7]++;
            }
            if (writePriv) {
               rep[2][7]++;
            }
            if (writeSet) {
               rep[3][7]++;
            }
            if (writeCycle) {
               rep[4][7]++;
            }
         }

         // Test if there is something to be written to external files
         if (!originalPath.isEmpty()) {
            if (writePrep) {
               rep[1][5]++;
               writeToFile (prepFile, oldTagStr, params.groupASpaths, params.showPrefixes,
                           params.alwaysPrefix, params.uniquePrefixes, tag,
                           parser.parsedData.network, parser.parsedData.netmask,
                           originalPath, pathSet[1]);
            }

            if (writeCycle) {
               rep[4][5]++;
               writeToFile (cycleFile, oldTagStr, params.groupASpaths, params.showPrefixes,
                           params.alwaysPrefix, params.uniquePrefixes, tag,
                           parser.parsedData.network, parser.parsedData.netmask,
                           originalPath, pathSet[2]);
            }

            if (writePriv) {
               rep[2][5]++;
               writeToFile (privFile, oldTagStr, params.groupASpaths, params.showPrefixes,
                           params.alwaysPrefix, params.uniquePrefixes, tag,
                           parser.parsedData.network, parser.parsedData.netmask,
                           originalPath, pathSet[3]);
            }

            if (writeSet) {
               rep[3][5]++;
               writeToFile (setFile, oldTagStr, params.groupASpaths, params.showPrefixes,
                           params.alwaysPrefix, params.uniquePrefixes, tag,
                           parser.parsedData.network, parser.parsedData.netmask,
                           originalPath, pathSet[4]);
            }

            if (writeIncomp) {
               rep[0][5]++;
               writeToFile (incompFile, oldTagStr, params.groupASpaths, params.showPrefixes,
                           params.alwaysPrefix, params.uniquePrefixes, tag,
                           parser.parsedData.network, parser.parsedData.netmask,
                           originalPath, pathSet[5]);
            }
         }
         else {
            // Just keep trace of discarded empty paths
            rep[5][6]++;
         }
      } // parser.parseLine() == PARSE_NOERROR
   }

   if (showPercentage) {
      // Print percentage of completion
      printPercentage((float) parser.filePos(), (float) parser.fileSize());
   }

   /* If user asked for path aggregation, results cannot be printed out
      until now */
   if (params.groupASpaths) {
      if (showPercentage) {
         printf ("\nNow writing output to file...\n");
      }
      printMap (outFile, pathSet[0], tag, showPercentage, params.showPrefixes, params.alwaysPrefix);
      if (prepFile != NULL && !pathSet[1].isEmpty()) {
         if (showPercentage) {
            printf ("\nNow writing prepended paths to file...\n");
         }
         printMap (prepFile, pathSet[1], tag, showPercentage, params.showPrefixes, params.alwaysPrefix);
      }
      if (cycleFile != NULL && !pathSet[2].isEmpty()) {
         if (showPercentage) {
            printf ("\nNow writing paths with cycles to file...\n");
         }
         printMap (cycleFile, pathSet[2], tag, showPercentage, params.showPrefixes, params.alwaysPrefix);
      }
      if (privFile != NULL && !pathSet[3].isEmpty()) {
         if (showPercentage) {
            printf ("\nNow writing paths with private ASs to file...\n");
         }
         printMap (privFile, pathSet[3], tag, showPercentage, params.showPrefixes, params.alwaysPrefix);
      }
      if (setFile != NULL && !pathSet[4].isEmpty()) {
         if (showPercentage) {
            printf ("\nNow writing paths with AS sets to file...\n");
         }
         printMap (setFile, pathSet[4], tag, showPercentage, params.showPrefixes, params.alwaysPrefix);
      }
      if (incompFile != NULL && !pathSet[5].isEmpty()) {
         if (showPercentage) {
            printf ("\nNow writing incomplete paths to file...\n");
         }
         printMap (incompFile, pathSet[5], tag, showPercentage, params.showPrefixes, params.alwaysPrefix);
      }
   }

   if (params.verbose) {
      fprintf (stderr, "\n\nAS paths     unchanged     changed     deleted       saved      tagged\n");
      fprintf (stderr, "incomplete %11lu %11s %11lu %11lu %11lu\n",
         rep[0][0], "-", rep[0][2], rep[0][3], rep[0][4]);
      fprintf (stderr, "w/ prep.   %11lu %11lu %11lu %11lu %11lu\n",
         rep[1][0], rep[1][1], rep[1][2], rep[1][3], rep[1][4]);
      fprintf (stderr, "w/ private %11lu %11lu %11lu %11lu %11lu\n",
         rep[2][0], rep[2][1], rep[2][2], rep[2][3], rep[2][4]);
      fprintf (stderr, "w/ AS sets %11s %11lu %11lu %11lu %11lu\n",
         "-", rep[3][1], rep[3][2], rep[3][3], rep[3][4]);
      fprintf (stderr, "w/ cycles  %11lu %11lu %11lu %11lu %11lu\n",
         rep[4][0], rep[4][1], rep[4][2], rep[4][3], rep[4][4]);
      fprintf (stderr, "TOTAL      %11lu %11lu %11lu %11lu %11lu\n\n",
         rep[0][0] + rep[1][0] + rep[2][0] + rep[3][0] + rep[4][0],
         rep[0][1] + rep[1][1] + rep[2][1] + rep[3][1] + rep[4][1],
         rep[0][2] + rep[1][2] + rep[2][2] + rep[3][2] + rep[4][2],
         rep[0][3] + rep[1][3] + rep[2][3] + rep[3][3] + rep[4][3],
         rep[0][4] + rep[1][4] + rep[2][4] + rep[3][4] + rep[4][4]);
      fprintf (stderr, "                                         empty       empty\n");
      fprintf (stderr, "                                        before       after\n");
      fprintf (stderr, "AS paths     processed    reported  processing  processing\n");
      fprintf (stderr, "input file %11lu %11lu %11lu %11lu\n",
         correctLines, rep[5][5], rep[5][6], rep[5][7]);
      fprintf (stderr, "incom.file %11s %11lu %11s %11lu\n",
         "-", rep[0][5], "-", rep[0][7]);
      fprintf (stderr, "prep. file %11s %11lu %11s %11lu\n",
         "-", rep[1][5], "-", rep[1][7]);
      fprintf (stderr, "priv. file %11s %11lu %11s %11lu\n",
         "-", rep[2][5], "-", rep[2][7]);
      fprintf (stderr, "sets file  %11s %11lu %11s %11lu\n",
         "-", rep[3][5], "-", rep[3][7]);
      fprintf (stderr, "cycle file %11s %11lu %11s %11lu\n",
         "-", rep[4][5], "-", rep[4][7]);
      fprintf (stderr, "TOTAL      %11s %11lu\n\n",
         "", rep[0][5] + rep[1][5] + rep[2][5] + rep[3][5] + rep[4][5] + rep[5][5]);
      fprintf (stderr, "File lines        read      parsed   w/ errors\n");
      fprintf (stderr, "           %11lu %11lu %11lu\n",
         parser.readLines(), parser.parsedLines(), wrongLines);
   }

   if (stdoutIsTerm && stdinIsTerm) {
      printf ("\n");
   }
   parser.closeFile();

   return 0;
}
